module IntEnumerator where

import Data.ExtendedReal
import Data.Universe
import Debug.Trace
import Data.List as List
import Control.Exception(assert)

-- type Enumerator = [ Integer ] -- change to stream? 

type IntEnumerator =
  (Extended Integer, Extended Integer) -- lower bound, upper bound

showIntEnum enum=
  case enum of 
    (NegInf, PosInf) -> "ℤ"
    (Finite k, PosInf) -> "["++(show k)++"..)"
    (NegInf, Finite k) -> "(.."++(show k)++"]"
    (Finite k, Finite l) -> "["++(show k)++".."++(show l)++"]"

elem i enum =
  case enum of 
    (NegInf, PosInf) -> True
    (Finite k, PosInf) -> k<=i
    (NegInf, Finite k) -> i<=k
    (Finite k, Finite l) -> k<=i && i<=l

intEnumerator lb ub =
  case lb of
    Nothing ->
      case ub of
        Nothing -> unboundedEnumerator
        Just ub -> upperBound ub
    Just lb ->
      case ub of
        Nothing -> lowerBound lb
        Just ub -> intDomain lb ub

unboundedEnumerator =
  (NegInf, PosInf)

subset:: IntEnumerator -> IntEnumerator -> Bool
subset (l1, u1) (l2, u2) =
  l1>=l2 && u1<=u2

intDomain:: Integer -> Integer -> IntEnumerator
intDomain k l = (Finite k, Finite l)

singleton::Integer -> IntEnumerator
singleton k = intDomain k k

upperBound::Integer -> IntEnumerator
upperBound k = (NegInf, Finite k)

lowerBound::Integer -> IntEnumerator
lowerBound k = (Finite k, PosInf)

union::IntEnumerator -> IntEnumerator -> IntEnumerator
union (lo1,hi1) (lo2,hi2) = ((min lo1 lo2), (max hi1 hi2))

intersection::IntEnumerator -> IntEnumerator -> IntEnumerator
intersection (lo1,hi1) (lo2,hi2) = ((max lo1 lo2), (min hi1 hi2))

firstVal::IntEnumerator -> Maybe Integer
firstVal iv | (isEmpty iv) = Nothing
firstVal (Finite k, PosInf) | k>=0 = Just k
firstVal (NegInf, Finite k) | k<=0 = Just k
firstVal (Finite lo, Finite hi) | hi<=0 = Just hi
firstVal (Finite lo, Finite hi) | lo>=0 = Just lo
firstVal iv = assert (IntEnumerator.elem 0 iv) (Just 0)

{- nextValue in enumerator according to enumerating order (NOT according to usual mathematical order!) -}
nextVal::IntEnumerator -> Integer -> Maybe Integer
nextVal (NegInf, PosInf) v =
  case v of
    0 -> Just 1
    v | v>0 -> Just (-v)
    v | v<0 -> Just (-v+1)
nextVal intEnum v =
  case nextVal (NegInf, PosInf) v of
    Nothing -> Nothing
    Just nV -> 
      if IntEnumerator.elem nV intEnum then Just nV
      else
        if (Finite (abs nV))<(minNorm intEnum) then firstVal intEnum
        else if (Finite (abs nV))>(maxNorm intEnum) then Nothing
        else nextVal intEnum nV
    
index::Integer -> Integer
index k =
  if k>0 then 2*k-1
  else 2*(-k)

before::IntEnumerator {- nonempty -} -> Integer -> Bool 
before (NegInf, PosInf) l = False
before (Finite k, PosInf) l = (index l)<(index k)
before (NegInf, Finite k) l = (index l)<(index k)
before (Finite k1, Finite k2) l =
  if k1<=0 && 0<=k2 then False
  else if 0<=k1 && k1<=k2 then (index l)<(index k1)
  else if k1<=k2 && k2<=0 then (index l)<(index k2)
  else assert False False

stream:: IntEnumerator -> [ Integer ]        
stream (NegInf, PosInf) = universe
stream (Finite k, PosInf) = [k..]
stream (NegInf, Finite k) = List.map negate [k..]
stream (Finite lo, Finite hi)= [lo..hi] 

choose:: IntEnumerator -> Integer 
choose (NegInf, PosInf) = 0
choose (Finite k, PosInf) = k
choose (NegInf, Finite k) = k
choose (Finite lo, Finite hi) | lo<=hi = lo

fixesValue:: IntEnumerator -> Maybe Integer
fixesValue (Finite k, Finite l) | k==l = Just k
fixesValue _ = Nothing

isFinite:: IntEnumerator -> Bool
isFinite (Finite _, Finite _) = True
isFinite _ = False

isEmpty:: IntEnumerator -> Bool
isEmpty (Finite k, Finite l) | k>l = True
isEmpty _ = False

minNorm::IntEnumerator -> Extended Integer
minNorm (NegInf, PosInf) = 0
minNorm (Finite k, PosInf)= if k<0 then Finite (-k) else 0
minNorm (NegInf, Finite k)= if k<0 then Finite (-k) else 0
minNorm (Finite lo, Finite hi)=
  if lo>hi then PosInf
  else if lo<0 && 0<hi then Finite 0
  else Finite (min (-lo) (hi))

maxNorm::IntEnumerator -> Extended Integer
maxNorm (NegInf, PosInf) = PosInf
maxNorm (Finite _, PosInf)= PosInf
maxNorm (NegInf, Finite _)= PosInf
maxNorm (Finite lo, Finite hi)=
  if lo>hi then NegInf
  else Finite (max (abs lo) (abs hi))

normFirst:: IntEnumerator -> Integer -> Bool -> Maybe Integer 
normFirst enum norm lessEq =
  if lessEq then IntEnumerator.firstVal enum
  else {- equal -}
    if IntEnumerator.elem norm enum then Just norm
    else if IntEnumerator.elem (-norm) enum then Just (-norm)
    else Nothing
