module Spec(parseSpecification,showSpecification,sat,smtLib,SatResult(..)) where

import Data.ExtendedReal
import Text.Parsec
import Text.Parsec.String
import Text.Parsec.Expr
import qualified Text.Parsec.Token as P
import qualified Text.Parsec.Combinator as C
import Text.Parsec.Language (haskellDef)
import Data.Map.Strict as Map
import Data.List as List
import Data.Maybe
import Data.Set as Set
import Debug.Trace
import Control.Exception(assert)
import IntEnumerator
import Data.Foldable
import Type
import DataEntity
import VarVal
import VarAss
import Conflict
import PredSym
import ExecResult

trL level str obj =
  case level of
    0 -> obj {- trace str obj -}
    1 -> obj {- trace str obj -}
    2 -> obj {- trace str obj -}

data Term =
  Neg Term |
  Add Term Term |
  Mult Term Term |
  Constant Integer |
  Variable VariableName |
  FuncApp VariableName Term deriving (Eq)

showTerm t=
   case t of
     Neg t -> "-"++(showTerm t)
     Add t1 (Neg t2) -> (showTerm t1)++"-"++(showTerm t2)
     Add t1 t2 -> (showTerm t1)++"+"++(showTerm t2)
     Mult t1 t2 -> (showTerm t1)++"*"++(showTerm t2)
     Constant k -> (show k)
     Variable n -> n
     FuncApp n t -> n++"("++(showTerm t)++")"

instance Show Term where
  show = showTerm

diffTerm t1 t2 = Add t1 (Neg t2)

data QuantType = Existential | Universal deriving (Eq)
showQuantType Existential = "∃"
showQuantType Universal = "∀"

instance Show QuantType where
  show = showQuantType

quantToBool Existential = True
quantToBool Universal = False

type TypeDecl = (VariableName, Type)
type DeclBlock = [ TypeDecl ] 

parseInteger:: Parsec String st ()
parseInteger =
  do (string "ℤ") <|> (string "integer"); return ()

parseFuncArrow:: Parsec String st ()
parseFuncArrow =
 do (string "→" <|> string "->"); return ()

parseType:: Parsec String st Type
parseType =
  do
    parseDomain;
    option TypeInteger (do parseFuncArrow; parseInteger; return TypeFunction)

showDeclBlock declBlock= List.intercalate ", " (List.map showTypeDecl declBlock)
showTypeDecl (v, t)= v++":"++(showType t)

{-
arrayVars:: DeclBlock -> Set VariableName
arrayVars declBlock=
  let arrays= (List.filter (\(n, t)-> t==TypeFunction) declBlock) in
  Set.fromList (List.map fst arrays)
-}


data Spec =
  BoolConst Bool |
  Pred Term PredSym |
  Conjunction [ Spec ] |
  Disjunction [ Spec ] |
  Not Spec |
  Quant QuantType DeclBlock Spec deriving (Eq)

instance Show Spec where
  show (BoolConst b) = show b
{-  show (TypeDecl varName t) = varName++":"++(show t) -}
  show (Not s)= "¬ ["++(show s)++"]"
  show (Conjunction (h:t)) = "["++(List.foldl' (\s c -> s++" ∧ "++ (show c)) (show h) t)++"]"
  show (Disjunction [(Not s1),s2]) = "["++(show s1)++"] ⇒ "++ (show s2)
  show (Disjunction (h:t)) = "["++(List.foldl' (\s c -> s++" ∨ "++ (show c)) (show h) t)++"]"
  show (Pred (Add t1 (Neg t2)) predSym) = (show t1)++(show predSym)++(show t2)
  show (Pred term predSym) = (show term)++(show predSym)++"0"
  show (Quant q decl s) = (show q)++" "++(showDeclBlock decl)++" . "++"["++(show s)++"]"

type Specification = (DeclBlock, Spec)

showSpecification (d, s) = (showDeclBlock d)++ " . " ++ (show s)

parseSpecification:: Parsec String st Specification
parseSpecification = 
  do
    d<- parseDeclBlock
    spaces
    P.dot lexer
    spaces
    p<- parseSpec
    eof
    return (d, p)

lexer       = P.makeTokenParser haskellDef -- is this o.k.?

parseVariableName:: Parsec String st VariableName
parseVariableName = P.identifier lexer

parseIntegerString:: Parsec String st String
parseIntegerString =
  (string "ℤ" <|> string "integer")

--parseIntegerType =
--  do parseIntegerString; space; return TypeInteger

parseClosedInterval:: Parsec String st (Term,Term)
parseClosedInterval =
  (P.brackets lexer) (do lb<-parseExpr; string ".."; ub<- parseExpr; return (lb, ub))

parseInterval:: Parsec String st (Maybe Term, Maybe Term)
parseInterval =
  let parseOpenLb = do string "("; return Nothing in
  let parseClosedLb = do string "["; lb<- parseExpr; return (Just lb) in
  let parseOpenUb = do string ")"; return Nothing in
  let parseClosedUb = do ub<-parseExpr; string "]"; return (Just ub) in
  do
    lb<- ((try parseOpenLb) <|> (try parseClosedLb))
    string ".."
    ub<- ((try parseOpenUb) <|> (try parseClosedUb))
    return (lb, ub)
      
parseDomain:: Parsec String st (Maybe Term, Maybe Term)
parseDomain =
  (do string "ℤ"; return (Nothing, Nothing)) <|> (do string "integer"; return (Nothing, Nothing)) <|> parseInterval

parseVariable:: Parsec String st Term
parseVariable =
  do
    varName <- parseVariableName
    funcApp <- optionMaybe (P.parens lexer parseExpr)
    return (if isNothing funcApp then Variable varName else FuncApp varName (fromJust funcApp))

parseNatural:: Parsec String st Term
parseNatural = Constant <$> (P.natural lexer)

parseExpr    = buildExpressionParser table parseTerm
--         <?> "expression"

parseTerm:: Parsec String st Term
parseTerm    =  parseNatural <|> parseVariable <|> (P.parens lexer parseExpr)
--         <?> "simple expression"

table   = [ [prefix "-" Neg, prefix "+" id ]
           , [binary "*" Mult AssocLeft ] 
           , [binary "+" Add AssocLeft, binary "-" diffTerm AssocLeft ]
           ]

binary  name fun assoc = Infix (do{ P.reservedOp lexer name; return fun }) assoc
prefix  name fun       = Prefix (do{ P.reservedOp lexer name; return fun })
postfix name fun       = Postfix (do{ P.reservedOp lexer name; return fun })

parseEqual::Parsec String st PredSym
parseEqual= do string "="; spaces; return Equal

parseNotEqual::Parsec String st PredSym
parseNotEqual= do (string "≠" <|> string "<>"); spaces; return NotEq

parseGreaterEq::Parsec String st PredSym
parseGreaterEq= do (string "≥" <|> string ">="); spaces; return GreaterEq

parseLessEq::Parsec String st PredSym
parseLessEq= do (string "≤" <|> string "<="); spaces; return LessEq

parseLess::Parsec String st PredSym
parseLess= do string "<"; spaces; return Less

parseGreater::Parsec String st PredSym
parseGreater= do string ">"; spaces; return Greater

parsePredSym =
  try parseEqual <|>
  try parseNotEqual <|>
  try parseGreaterEq <|> -- must be BEFORE Greater
  try parseGreater <|>
  try parseLessEq <|> -- must be BEFORE Less
  try parseLess 
  
parsePred:: Parsec String st Spec
parsePred=
  do
    term1 <- parseExpr
    predSym <- parsePredSym
    term2 <- parseExpr
    return
      (case term2 of
        Constant 0 -> Pred term1 predSym
        _ -> Pred (Add term1 (Neg term2)) predSym)

parseTypeDecl:: Parsec String st TypeDecl
parseTypeDecl=
  do
    varName <- parseVariableName
    P.colon lexer
    t <- parseType
    spaces
    return (varName, t)

parseDeclBlock:: Parsec String st DeclBlock
parseDeclBlock =
  C.sepBy parseTypeDecl (P.comma lexer)

buildQuant:: QuantType -> DeclBlock -> Spec -> Spec
buildQuant q decl s = Quant q decl s

parseUniversalQuant:: Parsec String st QuantType
parseUniversalQuant =
  do
    (P.symbol lexer "forall") <|> (P.symbol lexer "∀")
    return Universal

parseExistentialQuant:: Parsec String st QuantType
parseExistentialQuant =
  do 
    (P.symbol lexer "exists") <|> (P.symbol lexer "∃")
    return Existential

parseQuant:: Parsec String st Spec
parseQuant =
  do
    s<- parseUniversalQuant <|> parseExistentialQuant
    spaces
    v<- parseDeclBlock
    spaces
    char '.'
    spaces
    p<- parseSpec
    return (buildQuant s v p) 

exprLog    = buildExpressionParser tableLog termLog
--         <?> "expression"

termLog:: Parsec String st Spec
termLog    =
  try parsePred <|>
{-  try parseTypeDecl <|> -}
  try parseQuant <|>
  try (P.brackets lexer exprLog)    -- try is inefficient 
--         <?> "simple expression"

buildImplication:: Spec -> Spec -> Spec
buildImplication s_left s_right =
  Disjunction [(Not s_left), s_right]

buildConjunction:: Spec -> Spec -> Spec
buildConjunction (Conjunction s1) (Conjunction s2) = Conjunction (s1 ++ s2)
buildConjunction s1 (Conjunction s2) = Conjunction (s1:s2)
buildConjunction (Conjunction s1) s2 = Conjunction (s1 ++ [ s2 ])
buildConjunction s1 s2 = Conjunction [ s1, s2 ]

buildDisjunction:: Spec -> Spec -> Spec
buildDisjunction (Disjunction s1) (Disjunction s2) = Disjunction (s1 ++ s2)
buildDisjunction s1 (Disjunction s2) = Disjunction (s1:s2)
buildDisjunction (Disjunction s1) s2 = Disjunction (s1 ++ [ s2 ])
buildDisjunction s1 s2 = Disjunction [ s1, s2 ]


tableLog   = [
            [prefix "not" Not, prefix "¬" Not ],
            [binary "==>" buildImplication AssocLeft, binary "⇒" buildImplication AssocLeft],
            [binary "/\\" buildConjunction AssocLeft, binary "∧" buildConjunction AssocLeft, binary "\\/" buildDisjunction AssocLeft, binary "∨" buildDisjunction AssocLeft ]
           ]


parseSpec:: Parsec String st Spec
parseSpec =
  exprLog

type EvalRes = (Integer, Set DataEntity)

combEvalRes op (k1, v1) (k2, v2) = ((op k1 k2), Set.union v1 v2)

evalTerm:: Term -> VarAss -> EvalRes 
evalTerm (Add t1 t2) v = combEvalRes (+) (evalTerm t1 v) (evalTerm t2 v)
evalTerm (Neg t) v = let (k, vars)= evalTerm t v in (-k, vars)
evalTerm (Mult t1 t2) v = combEvalRes (*) (evalTerm t1 v) (evalTerm t2 v)
evalTerm (Constant k) v = (k, Set.empty)
evalTerm (Variable name) v =
  (VarAss.lookup v (Integer name), Set.singleton (Integer name))
evalTerm (FuncApp name term) v =
  let (arg_val, arg_vars)= evalTerm term v in
  let entity= (Function name arg_val) in
  (VarAss.lookup v entity, Set.insert entity arg_vars)


data SatResult =
{-   SyntaxError | -}
  Sat VarAss |
  Unknown |
  UnSat -- deriving (Eq)

showSatResult r=
  case r of
    Sat varAss -> "satisfiable\n"++(showVarAss varAss)
    UnSat -> "unsatisfiable\n"
    Unknown -> "unknown\n"

instance Show SatResult where
  show = showSatResult


equiv True True = True
equiv False False = True
equiv True False = False
equiv False True = False

execConj:: [Spec] -> Bool -> VarAss -> ExecResult
execConj s b v =
  foldl1 conj (List.map (\s -> exec s b v) s)

execDisj:: [Spec] -> Bool -> VarAss -> ExecResult
execDisj s b v =
  foldl1 disj (List.map (\s -> exec s b v) s)

{-
newInfo varAss conflict =
  let isEq (d, predSym, k) = predSym==Equal in
  let isNew (d, _, _)= not (VarAss.member varAss d) in 
  not (List.all isEq conflict) || (List.any isNew conflict)
-}
 
execExistsBounds::(VarAss -> ExecResult) -> VarAss -> VarAssEnum -> ExecResult
execExistsBounds exec v tries = 
  let currVarAss= currEnum (assert (isConsistent tries) tries) in
  let varAssExt= varAssExtend v currVarAss in 
  case exec (trL 0 ("execVarAssExt: "++(show varAssExt)) varAssExt) of
    r@(Just (False, conflict, _))->
      case addConstraints tries (trL 2 ("conflict: "++(show conflict)) conflict) of
        Just updatedEnum ->
          let nextEnum= nextEnumOK (trL 2 ("updatedEnum: "++(show updatedEnum)++" from "++(show tries)) updatedEnum) in
          case trL 2 ("nextEnum: "++(show nextEnum)) nextEnum of    
            Nothing ->  execResultUnsat 
            Just next -> disj r (execExistsBounds exec v next)
        Nothing -> execResultUnsat
    (Just (True, conflict, _))->
      Just (True, conflict, currVarAss)  

isDeclared:: DeclBlock -> BasicConstraint -> Bool
isDeclared decl (entity, _, _) =
  List.any (\(v, t) -> v==(DataEntity.varName entity)) decl 


execExistsLearn::DeclBlock -> (VarAss -> ExecResult) -> VarAss -> EnumList -> ExecResult
execExistsLearn decl exec_f v enum =
  let init= initEnumNormOK enum decl in
  if isNothing init then execResultUnsat
  else 
    case execExistsBounds exec_f v (fromJust init) of
      Just (False, conflict, witness) | not (Conflict.isEmpty conflict) ->
        let nonLocalConflict = Conflict.filter (not . (isDeclared decl)) conflict in
        Just (False, trL 2 ("nonLocalConflict"++(show decl)++" "++(show nonLocalConflict)) nonLocalConflict, witness)
      Just (True, conflict, witness) | not (Conflict.isEmpty conflict) ->
        Just (True, Conflict.filter (not . (isDeclared decl)) conflict, witness)
      r@_ -> (trL 2 ("execExistsLearnRes: "++(show r)) r)
      
execExists:: Specification -> Bool -> VarAss -> ExecResult
execExists (decl, spec) b v =
  let exec_f= exec spec b in
  let enum= enumListFromTypeDecl decl in
  let r= execExistsLearn decl exec_f v enum in
  trL 2 ("execExistsRes: "++(show r)) r
    
execForall:: Specification -> Bool -> VarAss -> ExecResult
execForall s b v =
  ExecResult.negate (execExists s (not b) v)

generalize:: VarAss -> DataEntity-> Integer -> PredSym -> Conflict
generalize varAss d k predSym =
  conflict d predSym k

pruneEqual:: DataEntity -> VarAss -> Integer -> ExecResult
pruneEqual dataEntity v evalRes =
  let val= VarAss.lookup v dataEntity in
    if val==evalRes then Just (True, emptyConflict, emptyVarAss)
    else
      let gen=
            if val>evalRes then conflict dataEntity GreaterEq  (evalRes+1)
            else conflict dataEntity LessEq (evalRes-1) in
        Just (False, gen, emptyVarAss)

execPrune dataEntity v val vars predSymAdj=
  let curVal= VarAss.lookup v dataEntity in
  let r= (predOp predSymAdj) curVal val in
  let c= conflictFromVarAss v vars in
  let conflict = Conflict.add c (dataEntity, if r then predSymAdj else negPred predSymAdj,val)  in
{-  execResult (tr ((show predSymAdj)++" "++(show curVal)++" "++(show val)++" "++(show r)) r) conflict -}
  execResult r conflict

execPred:: Term -> PredSym -> Bool -> VarAss -> ExecResult
execPred t predSym b v =
  let predSymAdj= if b then predSym else negPred predSym in
  case t of
    Variable varName ->
      let (val, vars)= (0, Set.empty) in
      let dataEntity= Integer varName in          
      execPrune dataEntity v val vars predSymAdj
    Add (Variable varName) (Neg t) ->
      let (val, vars)= evalTerm t v in
      let dataEntity= Integer varName in                  
      execPrune dataEntity v val vars predSymAdj        
    FuncApp varName arg  ->    
      let (val, vars)= (0, Set.empty) in
      let (valArg, varsArg)= evalTerm arg v in        
      let dataEntity= (Function varName valArg) in
      execPrune dataEntity v val (Set.union vars varsArg) predSymAdj
    Add (FuncApp varName arg) (Neg t) -> 
      let (val, vars)= evalTerm t v in
      let (valArg, varsArg)= evalTerm arg v in        
      let dataEntity= (Function varName valArg) in
      let allVars= Set.union vars varsArg in
      execPrune dataEntity v val allVars predSymAdj
    _ -> 
     let (val, vars)= evalTerm t v in
     let conflict = conflictFromVarAss v vars in
     execResult (predOp predSymAdj val 0) (trL 2 ("evalConf:"++(show conflict)) conflict)

exec:: Spec -> Bool -> VarAss -> ExecResult
exec s b v =
  case trL 2 ("exec: "++(show s)++" "++(show b)++" "++(show v)) s  of  
--  case s of
    BoolConst c -> if equiv b c then (execResult True emptyConflict) else execResultUnsat    
    Pred t predSym ->
      let r= execPred t predSym b v in
      trL 2 ("resPred: "++(show r)) r 
    Not s ->
      let r= exec s (not b) v in
      trL 2 ("resNot: "++(show v)++" "++(show r)) r
    Conjunction s -> if b then execConj s b v else execDisj s b v 
    Disjunction s -> if b then execDisj s b v else execConj s b v
    Quant q decl s ->
      let qb= quantToBool q in
      if qb==b then execExists (decl, s) b v else execForall (decl, s) b v

type VarAssEnum= (EnumList, DeclBlock {- variables declared locally (current enumeration/quantifier) -}, Integer {- current norm -}, [ Integer ])
 {- invariant: isConsistent (see next function) -} 
 {- make instance of Control.Monad.State? -}

equalState:: VarAssEnum -> VarAssEnum -> Bool
equalState (_, _, _, l1) (_, _, _, l2) = l1==l2

isConsistent:: VarAssEnum -> Bool
isConsistent v@(enum, _, _, current)= 
  {- TODO: also check for correct norm -}
  (isConsistentStrict v) && (isConsistentVal (List.map snd enum) current)

isConsistentVal enum current =
  let consVal (e, c) = IntEnumerator.elem c e in
  List.all consVal (zip enum current)

isConsistentStrict:: VarAssEnum -> Bool
isConsistentStrict (enumList, _, _, state)=
  (List.length enumList)==(List.length state)

makeVarAssEnum e arr n l =
  let res= (e, arr, n, l) in
  assert (isConsistentStrict res) res

addConstraints:: VarAssEnum -> Conflict -> Maybe VarAssEnum 
addConstraints v conflict =
  let (enum, locVars, norm, current) = addEntities v conflict in
  let localConflicts=
        let isLocal (entity, _,_) = List.any (\(n,_)-> (varName entity)==n) locVars in
        Conflict.filter isLocal conflict in
  let consistentConflict= Conflict.filter (isConsistentBC enum) localConflicts in
  case Conflict.size (trL 2 ("consistentConflicts: "++(show (consistentConflict,v))) consistentConflict) of
    1 -> Just (makeVarAssEnum (Conflict.foldl addConstraint enum consistentConflict) locVars norm current)
    0 -> Nothing
    r@_ -> Just (makeVarAssEnum enum locVars norm current)  

makeConsistent:: VarAssEnum -> Maybe VarAssEnum     
makeConsistent v@(enumList, arr, norm, l) =
  case makeConsistentNorm (assert (isConsistentStrict v) v) of {- make consistent using the current norm -} 
    Nothing ->
      if (Finite norm)>normUpperBound enumList then
        trL 1 ("covered: "++(show norm)++" "++(show enumList)) Nothing
      else
        trL 1 ("increase enumeration norm "++(show (norm+1))++" "++(show enumList)) makeConsistent (enumList, arr, norm+1, initEnumList (List.length enumList)) {- replace +1 by normLowerBound -}
    Just r ->
      case r of (l1,_,_,l2) -> Just r

{-
covered:: EnumList -> Integer -> Bool
covered enumList n =
  let covN (_, enum) = IntEnumerator.subset enum ((Finite (-n)), (Finite n)) in
-}

 
makeConsistentNorm:: VarAssEnum -> Maybe VarAssEnum
makeConsistentNorm (enumList, arr, norm, l) =
  case makeConsistentNormList enumList norm l False [] [] of
    Nothing -> Nothing
    Just l ->
      let v = makeVarAssEnum enumList arr norm l in
      if (List.elem norm l) || (List.elem (-norm) l) then Just v
      else makeConsistentNorm (nextEnum v)
      


nextEnum:: VarAssEnum -> VarAssEnum
nextEnum (enumList, locVars, norm, l)=
  (enumList, locVars, norm, nextIntList l)

nextEnumOK:: VarAssEnum -> Maybe VarAssEnum
nextEnumOK v =
  let next= nextEnum v in
  let r= makeConsistent (trL 2 ("next:"++(show next)) next) in
  case r of
    Nothing -> Nothing
    Just r -> assert (not (equalState v r)) (Just r)

  
currEnum:: VarAssEnum -> VarAss
currEnum c@(enum, _, _, l) =  
  toAss (List.map fst enum) (assert (isConsistent c) l)

sat:: Specification -> SatResult
sat s =
  case execExists s True VarAss.emptyVarAss of
    Just (True, _, w) -> Sat w
    Just (False, _, _) -> UnSat
    _ -> Unknown

data SpecSmtLib = SmtLib Specification

instance Show SpecSmtLib where
  show (SmtLib spec)= showSmtLibSpecification spec

smtLib spec= SmtLib spec

showSmtLibTerm (Neg term)= "(- "++(showSmtLibTerm term)++")"
showSmtLibTerm (Add t1 t2)= "(+ "++(showSmtLibTerm t1)++" "++(showSmtLibTerm t2)++")"
showSmtLibTerm (Mult t1 t2)= "(* "++(showSmtLibTerm t1)++" "++(showSmtLibTerm t2)++")"
showSmtLibTerm (Constant k)= (show k)
showSmtLibTerm (Variable name)= name
showSmtLibTerm (FuncApp name t)= "("++name++" "++(showSmtLibTerm t)++")"

concatBlank s1 s2 = s1 ++ " " ++ s2
  
showSmtLibSpec:: Spec -> String
showSmtLibSpec (BoolConst b) = ""
showSmtLibSpec (Pred t predSym) = "("++(showPredSym False predSym)++" "++(showSmtLibTerm t)++" 0)"
showSmtLibSpec (Conjunction s) = "(and "++(List.foldl1' concatBlank (List.map showSmtLibSpec s))++")"
showSmtLibSpec (Disjunction s) = "(or "++(List.foldl1' concatBlank (List.map showSmtLibSpec s))++")"
showSmtLibSpec (Not s) = "(not "++(showSmtLibSpec s)++") "
showSmtLibSpec (Quant q [(varName,typeDecl)] s) =
  "("++(if q==Existential then "exists " else "forall ")++
  "(("++varName++" "++(showSmtLibType typeDecl)++")) "++
  (showSmtLibSpec s)++
  ") "

showSmtLibDecl::TypeDecl -> String
showSmtLibDecl (v, t) = 
  case t of
    TypeInteger -> "(declare-const "++v++" Int)\n" 
    TypeFunction -> "(declare-fun "++v++" (Int) Int)\n"

showSmtLibDeclBlock:: DeclBlock -> String
showSmtLibDeclBlock declBlock= List.foldl (++) "" (List.map showSmtLibDecl declBlock)

showSmtLibSpecification:: Specification -> String
showSmtLibSpecification (declBlock, spec) =
  (showSmtLibDeclBlock declBlock)++
  "(assert "++(showSmtLibSpec spec)++")\n"++
  "(check-sat)\n"++
  "(get-model)\n"
  
extractType:: Spec -> (Spec, Map VariableName Type)
extractType spec =
  case spec of
{-    TypeDecl name t -> (BoolConst True, Map.singleton name t) -}
    Conjunction s ->
      let (specs, typeList)= List.unzip (List.map extractType s) in
      let types= Map.unions typeList in 
      if List.any (\s -> s==(BoolConst False)) specs then (BoolConst False, types)
      else
        case List.filter (\s -> not (s==BoolConst True)) specs of
          [] -> (BoolConst True, types)
          l -> (Conjunction l, types)
    Disjunction s ->
      let (specs, typeList)= List.unzip (List.map extractType s) in
      let types= Map.unions typeList in 
      if List.any (\s -> s==(BoolConst True)) specs then (BoolConst True, types)
      else
        case List.filter (\s -> not (s==BoolConst False)) specs of
          [] -> (BoolConst False, types)
          l -> (Disjunction l, types)
    (Not spec) ->
      let (s, t)= extractType spec in
      (Not s, t)  
    s -> (s, Map.empty)  


showSmtLibType:: Type -> String
showSmtLibType TypeInteger= "Int"
showSmtLibType TypeFunction= "(Int) Int"
  
assertType:: VariableName -> Type -> String
assertType varName varType =
  case varType of
    t@(TypeInteger) ->
      "(declare-const "++varName++" "++(showSmtLibType t)++")\n"
    t@(TypeFunction) ->
      "(declare-fun "++varName++" "++(showSmtLibType t)++")\n"
    
showSmtLib spec =
  let (specNoT, types)= extractType spec in
  (Map.foldr (++) "" (Map.mapWithKey assertType types))++
  "(assert "++(showSmtLibSpec specNoT)++")\n"++
  "(check-sat)\n(get-model)\n"
  
isConsistentBC:: EnumList -> BasicConstraint -> Bool
isConsistentBC enum bc@(dataEntityBC,predSym,const) =
  let isInConsistentEnum (dataEntity, e) = dataEntity==dataEntityBC && implies e predSym const in
  not (List.any isInConsistentEnum enum)

addEntities:: VarAssEnum -> Conflict -> VarAssEnum
addEntities v@(enum, locVars, norm, current) conflict =
  let newEntities= Conflict.mapMaybe (newEntity locVars enum) conflict in
  case newEntities of
    [] -> v
    _ -> fromJust (initEnumNormOK (enum++newEntities) locVars)
{-      (enum++newEntities, locVars, norm, (current++(List.replicate (List.length newEntities) VarAss.defaultVal))) -}
  
newEntity:: DeclBlock-> EnumList -> BasicConstraint -> Maybe (DataEntity, IntEnumerator)
newEntity locVars enum (d, _,_) =
  let add dataEntity =
        case dataEntity of
          Function name _ | (List.elem (name,TypeFunction) locVars) && (not (List.elem dataEntity (fst (unzip enum)))) -> True
          _ -> False in
  if add d then Just (d, IntEnumerator.unboundedEnumerator) 
  else Nothing


update:: BasicConstraint -> (DataEntity, IntEnumerator) -> (DataEntity, IntEnumerator)
update bc (d, intEnum) =
  case bc of
    (e, _, _) | d==e ->
      let newEnum= IntEnumerator.intersection intEnum (enumFromNegBC bc) in {- intersection with negation! -}
      (d, newEnum)
    _ -> (d, intEnum)

addConstraint:: EnumList -> BasicConstraint -> EnumList
addConstraint enum bc@(dataEntity,_,_) = 
  List.map (Spec.update bc) enum     

type EnumList = [ (DataEntity, IntEnumerator) ]

enumListEmpty = []

enumListFromTypeDecl::DeclBlock -> EnumList
enumListFromTypeDecl declBlock =
  let intName d =
        case d of
          (name, TypeInteger) -> Just name
          _ -> Nothing in
  let intDeclList = catMaybes (List.map intName declBlock) in
  List.map (\n -> (Integer n, IntEnumerator.unboundedEnumerator)) intDeclList 

enumFromNegBC:: BasicConstraint -> IntEnumerator
enumFromNegBC (d, Equal, v) = IntEnumerator.unboundedEnumerator {- information loss -}
enumFromNegBC (d, GreaterEq, v) = IntEnumerator.upperBound (v-1)
enumFromNegBC (d, LessEq, v) = IntEnumerator.lowerBound (v+1)
enumFromNegBC (d, NotEq, v) = IntEnumerator.singleton v
enumFromNegBC (d, Greater, v) = IntEnumerator.upperBound v
enumFromNegBC (d, Less, v) = IntEnumerator.lowerBound v

enumFromConflict conflict@(d, pred, v)= (d, enumFromNegBC conflict)

initEnumList:: Int -> [ Integer ]
initEnumList dim = List.replicate dim defaultVal 

makeConsistentNormList:: EnumList -> Integer -> [ Integer ] -> Bool -> [ Integer ] -> [ Integer ] -> Maybe [ Integer ]  
makeConsistentNormList enumList norm l carry partRes {- reverse of partial result -} resetRes {- reverse of reset if carry -} =
  let t x = x {- trace (show (enumList,l,carry,partRes,resetRes)) x -} in  
  case assert ((List.length enumList)==(List.length l) && ((List.length partRes)==(List.length resetRes))) (t (enumList, l)) of
    ([],[]) ->
      if carry then Nothing
      else Just (reverse partRes)
    ((_,enumFirst):enumRest, h:t)->
      let enumInt= IntEnumerator.intersection enumFirst (Finite (-norm), (Finite norm)) in
      case IntEnumerator.firstVal enumInt of
        Nothing -> Nothing
        Just first -> 
          if not carry && (IntEnumerator.elem h enumInt) then
            makeConsistentNormList enumRest norm t False (h:partRes) (first:resetRes)
          else
            case IntEnumerator.nextVal enumInt h of
              Nothing -> makeConsistentNormList enumRest norm t True (first:resetRes) (first:resetRes)
              Just n -> makeConsistentNormList enumRest norm t False (n:resetRes) (first:resetRes)

initEnumNormOK:: EnumList -> DeclBlock -> Maybe VarAssEnum
initEnumNormOK enum locVars = initEnumNormOKN enum locVars (normLowerBound enum)
  
initEnumNormOKN:: EnumList -> DeclBlock -> Integer -> Maybe VarAssEnum
initEnumNormOKN l locVars n =
  makeConsistent (l, locVars, 0, initEnumList (List.length l))

nextIntList:: [ Integer ] -> [ Integer ]
nextIntList (f:r) = 
    case IntEnumerator.nextVal IntEnumerator.unboundedEnumerator f of
      Just nextVal -> nextVal:r
nextVal::IntEnumerator -> Integer -> Integer -> Maybe Integer    
nextVal intEnum n k =
  if k==n then Nothing
  else
    let r = IntEnumerator.nextVal intEnum k in
    r

norm:: [Integer] -> Integer
norm l =
  maximum l

normLowerBound:: EnumList -> Integer
normLowerBound enumList =
  let l= List.map (minNorm . snd) enumList in
  case List.foldl' max 0 l of
    Finite k -> k

normUpperBound:: EnumList -> Extended Integer
normUpperBound enumList =
  let enums= List.map snd enumList in
  if any IntEnumerator.isEmpty enums then NegInf
  else
    let l= List.map maxNorm enums in
    List.foldl' max 0 l

implies:: IntEnumerator -> PredSym -> Integer -> Bool
implies enum predSym const =
  case predSym of
    Equal ->
      case enum of
        (Finite k, Finite l) | k==const && l==const -> True
        _ -> False
    NotEq ->
      case enum of
        (_, Finite k) | k<const -> True
        (Finite k, _) | k>const -> True
        _ -> False
    LessEq ->
      case enum of
        (_, Finite k) | k<=const -> True
        _ -> False
    Less -> implies enum LessEq (const+1)
    GreaterEq ->
      case enum of
        (Finite k, _) | k>=const -> True
        _ -> False
    Greater -> implies enum GreaterEq (const-1)


createEnum:: PredSym -> Integer -> IntEnumerator
createEnum predSym k =
  case predSym of
    Equal -> IntEnumerator.singleton k
    GreaterEq -> IntEnumerator.lowerBound k
    LessEq -> IntEnumerator.upperBound k
    Greater -> IntEnumerator.lowerBound (k+1)
    Less -> IntEnumerator.upperBound (k-1)
    _ -> IntEnumerator.unboundedEnumerator
 
maxMaxNorm::EnumList -> Extended Integer
maxMaxNorm enumList =
  let uboundList= List.map (snd . snd) enumList in 
  List.foldl1' max uboundList
